// Copyright (c) 2016, Shinichiro Abe. All rights reserved. Use of this source code
// is governed by a BSD-style license that can be found in the LICENSE file.

abstract class Analyzer {
  List<CharFilter> charFilters = [];
  Tokenizer tokenizer;
  List<TokenFilter> tokenFilters = [];
  Iterable<String> getTokens(String text) {
    String normalized = text;
    for (var charFilter in charFilters) {
      normalized = charFilter.normalize(normalized);
    }
    Iterable<String> tokens = tokenizer.tokenize(normalized);
    Iterable<String> filteredTokens = tokens;
    for (var tokenFilter in tokenFilters) {
      filteredTokens = tokenFilter.filter(filteredTokens);
    }
    return filteredTokens;
  }
}
// ignore: one_member_abstracts
abstract class CharFilter {
  String normalize(String text);
}
// ignore: one_member_abstracts
abstract class Tokenizer {
  Iterable<String> tokenize(String text);
}
// ignore: one_member_abstracts
abstract class TokenFilter {
  Iterable<String> filter(Iterable<String> tokens);
}

abstract class PatternReplaceCharFilter implements CharFilter {
  Pattern pattern;
  String replace;
  @override String normalize(String text) => text.replaceAll(pattern, replace);
}
class PatternCharFilter extends PatternReplaceCharFilter {
  Pattern _pattern;
  String _replace;
  PatternCharFilter(this._pattern, this._replace) {
    pattern = _pattern;
    replace = _replace;
  }
}
class IdeographicSpaceCharFilter extends PatternReplaceCharFilter {
  IdeographicSpaceCharFilter() {
    pattern = new RegExp(r"　");
    replace = " ";
  }
}
class MappingCharFilter implements CharFilter {
  Map<String,String> _mapping;
  MappingCharFilter(this._mapping);
  @override String normalize(String text) {
    String txt = '';
    for (int i = 0; i < text.length; i++) {
      if (_mapping[text[i]] != null) {
        txt = txt + _mapping[text[i]];
      } else {
        txt = txt + text[i];
      }
    }
    return txt;
  }
}
class UpperFullWidthCharFilter extends MappingCharFilter {
  static final Map<String,String> _xMapping =
     {'Ａ':'A', 'Ｂ':'B', 'Ｃ':'C', 'Ｄ':'D', 'Ｅ':'E', 'Ｆ':'F', 'Ｇ':'G',
      'Ｈ':'H', 'Ｉ':'I', 'Ｊ':'J', 'Ｋ':'K', 'Ｌ':'L', 'Ｍ':'M', 'Ｎ':'N',
      'Ｏ':'O', 'Ｐ':'P', 'Ｑ':'Q', 'Ｒ':'R', 'Ｓ':'S', 'Ｔ':'T', 'Ｕ':'U',
      'Ｖ':'V', 'Ｗ':'W', 'Ｘ':'X', 'Ｙ':'Y', 'Ｚ':'Z'};
  UpperFullWidthCharFilter() : super(_xMapping);
}
class LowerFullWidthCharFilter extends MappingCharFilter {
  static final Map<String,String> _xMapping =
   {'ａ':'a', 'ｂ':'b', 'ｃ':'c', 'ｄ':'d', 'ｅ':'e', 'ｆ':'f', 'ｇ':'g',
    'ｈ':'h', 'ｉ':'i', 'ｊ':'j', 'ｋ':'k', 'ｌ':'l', 'ｍ':'m', 'ｎ':'n',
    'ｏ':'o', 'ｐ':'p', 'ｑ':'q', 'ｒ':'r', 'ｓ':'s', 'ｔ':'t', 'ｕ':'u',
    'ｖ':'v', 'ｗ':'w', 'ｘ':'x', 'ｙ':'y', 'ｚ':'z'};
  LowerFullWidthCharFilter() : super(_xMapping);
}
class LowerCaseCharFilter implements CharFilter {
  @override String normalize(String text) {
    return text.toLowerCase();
  }
}
class DigitSeparatorCharFilter implements CharFilter {
  @override String normalize(String text) {
    String txt = '';
    if (text.indexOf(',') == -1) return text;
    for (int i = 0; i < text.length; i++) {
      bool found = false;
      if (text[i] == ',') {
        int prevIndex = i - 1;
        if ((i != 0) && (StandardTokenizer.isDigit(text.codeUnitAt(prevIndex)))) {
          int nextIndex = i + 1;
          if (!(nextIndex > text.length - 1) && (StandardTokenizer.isDigit(text.codeUnitAt(nextIndex)))) {
            found = true;
          }
        }
      }
      if (!found) txt = txt + text[i];
    }
    return txt;
  }
}

class KeywordTokenizer implements Tokenizer {
  @override Iterable<String> tokenize(String text) => [text];
}
class WhitespaceTokenizer implements Tokenizer {
  final RegExpTokenizer regExpTokenizer
    = new RegExpTokenizer(new RegExp("[ ,.!?:;'\"\r\n]"));
  @override Iterable<String> tokenize(String text) {
    return regExpTokenizer.tokenize(text);
  }
}
class RegExpTokenizer implements Tokenizer {
  final RegExp _regExp;
  RegExpTokenizer(this._regExp);
  @override Iterable<String> tokenize(String text) {
    return text.split(_regExp).where((e) => e.length > 0);
  }
}
class NGramTokenizer implements Tokenizer {
  final int _n;
  NGramTokenizer(this._n);
  @override Iterable<String> tokenize(String text) {
    return new Iterable.generate(text.length, (i) {
      var endIndex = i + _n;
      if (!(endIndex > text.length)) {
        return text.substring(i, endIndex);
      }
    }).where((e) => e != null);
  }
}
class EdgeNGramTokenizer implements Tokenizer {
  final int _n;
  EdgeNGramTokenizer(this._n);
  @override Iterable<String> tokenize(String text) {
    return new Iterable.generate(_n, (i) {
      var endIndex = i + 1;
      if (!(endIndex > text.length)) {
        return text.substring(0, endIndex);
      }
    }).where((e) => e != null);
  }
}

class LowerCaseFilter implements TokenFilter {
  @override Iterable<String> filter(Iterable<String> tokens) {
    return tokens.map((e) => e.toLowerCase());
  }
}
class NGramFilter implements TokenFilter {
  final int _n;
  NGramFilter(this._n);
  @override Iterable<String> filter(Iterable<String> tokens) {
    return new Iterable.generate(tokens.length, (i) {
      var text = tokens.elementAt(i);
      return new Iterable.generate(text.length, (i) {
        var endIndex = i + _n;
        if (!(endIndex > text.length)) {
          return text.substring(i, endIndex);
        }
      }).where((e) => e != null);
    }).where((e) => e != null).expand((x) => x);
  }
}
class EdgeNGramFilter implements TokenFilter {
  final int _n;
  EdgeNGramFilter(this._n);
  @override Iterable<String> filter(Iterable<String> tokens) {
    return new Iterable.generate(tokens.length, (i) {
      var text = tokens.elementAt(i);
      return new Iterable.generate(_n, (i) {
        var endIndex = i + 1;
        if (!(endIndex > text.length)) {
          return text.substring(0, endIndex);
        }
      }).where((e) => e != null);
    }).where((e) => e != null).expand((x) => x);
  }
}
class ShingleFilter implements TokenFilter {
  final int _n;
  final String _separator;
  ShingleFilter(this._n, this._separator);
  @override Iterable<String> filter(Iterable<String> tokens) {
    return new Iterable.generate(tokens.length, (i) {
      var shingle = _doShingle(tokens, i, _n);
      return (shingle.length == _n) ? shingle.join(_separator) : null;
    }).where((e) => e != null);
  }
  List<String> _doShingle(Iterable<String> tokens, int i, int n) {
    var ret = <String>[];
    for (int x = i; x < i + n; x++) {
      if (!(x > tokens.length - 1)) {
        ret.add(tokens.elementAt(x));
      }
    }
    return ret;
  }
}
class SkipBiGramFilter implements TokenFilter {
  final int _k;
  final String _separator;
  SkipBiGramFilter(this._k, this._separator);
  @override Iterable<String> filter(Iterable<String> tokens) {
    return new Iterable.generate(tokens.length, (i) {
      return _doSkipBiGram(tokens, i, _k);
    }).where((e) => e != null).expand((x) => x);
  }
  List<String> _doSkipBiGram(Iterable<String> tokens, int i, int skip) {
    var ret = <String>[];
    int endIndex = i + skip + 1;
    for (int x = i; x < endIndex; x++) {
      int end = x + 1;
      if (!(end > tokens.length - 1)) {
        String str = tokens.elementAt(i) + _separator + tokens.elementAt(end);
        ret.add(str);
      }
    }
    return ret;
  }
}
class CompoundWordsFilter implements TokenFilter {
  final int max = 10;
  final List<String> _dictionary;
  CompoundWordsFilter(this._dictionary);
  @override Iterable<String> filter(Iterable<String> tokens) {
    return tokens.map((token) => _decompose(token)).expand((x) => x)
        .where((String e) => e.length > 0);
  }
  List<String> _decompose(String token) {
    var ret = <String>[];
    ret.add(token);
    for (int i = 0; i < token.length; i++) {
      for (int j = max - 1; j > 0; j--) {
        int endIndex = i + j + 1;
        if (!(endIndex > token.length)) {
          String text = token.substring(i, endIndex);
          if (text.length > 0 && _dictionary.contains(text)) {
            if (!ret.contains(text)) ret.add(text);
            break;
          }
        }
      }
    }
    return ret;
  }
}
abstract class StopFilter implements TokenFilter {
  List<String> stopWords = [];
  @override Iterable<String> filter(Iterable<String> tokens) {
    return tokens.where((e) => !stopWords.contains(e));
  }
}
class StopWordsFilter extends StopFilter {
  final List<String> _stopWords;
  StopWordsFilter(this._stopWords) {
    stopWords = _stopWords;
  }
}
class EnglishStopWordsFilter extends StopFilter {
  EnglishStopWordsFilter() {
    stopWords = [
      "a", "an", "and", "are", "as", "at", "be", "but", "by",
      "for", "if", "in", "into", "is", "it",
      "no", "not", "of", "on", "or", "such",
      "that", "the", "their", "then", "there", "these",
      "they", "this", "to", "was", "will", "with"];
  }
}
class JapaneseStopWordsFilter extends StopFilter {
  JapaneseStopWordsFilter() {
    stopWords = [
      'の','に','は','を','た','が','で','て','と','し','れ','さ',
      'ある','いる','も','する','から','な','こと','として',
      'い','や','れる','など','なっ','ない','この','ため','その',
      'あっ','よう','また','もの','という','あり','まで','られ',
      'なる','へ','か','だ','これ','によって','により','おり','より',
      'による','ず','なり','られる','において','ば','なかっ','なく',
      'しかし','について','せ','だっ','その後','できる','それ','う',
      'ので','なお','のみ','でき','き','つ','における','および','いう',
      'さらに','でも','ら','たり','その他','に関する','たち',
      'ます','ん','なら','に対して','特に','せる','及び',
      'これら','とき','では','にて','ほか','ながら','うち',
      'そして','とともに','ただし','かつて','それぞれ','または',
      'お','ほど','ものの','に対する','ほとんど','と共に','といった',
      'です','とも','ところ','ここ'];
  }
}
abstract class KeepFilter implements TokenFilter {
  List<String> keepWords = [];
  @override Iterable<String> filter(Iterable<String> tokens) {
    return tokens.where((e) => keepWords.contains(e));
  }
}
class KeepWordsFilter extends KeepFilter {
  final List<String> _keepWords;
  KeepWordsFilter(this._keepWords) {
    keepWords = _keepWords;
  }
}
abstract class OrderingFilter implements TokenFilter {
  bool reverse = false;
  @override Iterable<String> filter(Iterable<String> tokens) {
    return tokens.map((e) {
      List<String> str = new Iterable.generate(e.length, (i) => e[i]).toList();
      List<String> ordered = [];
      str.sort();
      ordered.addAll(str);
      if (reverse) {
        ordered = ordered.reversed.toList();
      }
      return ordered.join();
    });
  }
}
class OrderingNaturalFilter extends OrderingFilter {
  OrderingNaturalFilter() {
    reverse = false;
  }
}
class OrderingReverseFilter extends OrderingFilter {
  OrderingReverseFilter() {
    reverse = true;
  }
}
class ReverseStringFilter implements TokenFilter {
  @override Iterable<String> filter(Iterable<String> tokens) {
    return tokens.map((e) {
      List<String> str = new Iterable.generate(e.length, (i) => e[i]).toList();
      List<String> ordered = str.reversed.toList();
      return ordered.join();
    });
  }
}
class LengthFilter implements TokenFilter {
  final int _min;
  final int _max;
  LengthFilter(this._min, this._max);
  @override Iterable<String> filter(Iterable<String> tokens) {
    return tokens.where((e) => (_min <= e.length && e.length <= _max));
  }
}
class TypoToleranceFilter implements TokenFilter {
  final int _n;
  NGramFilter _nGram;
  final List<OrderingFilter> _orders
    = [new OrderingNaturalFilter(), new OrderingReverseFilter()];
  TypoToleranceFilter(this._n) {
    _nGram = new NGramFilter(_n);
  }
  @override Iterable<String> filter(Iterable<String> tokens) {
    var ite = new Iterable.generate(_orders.length, (i) {
      var order = _orders[i];
      return order.filter(tokens);
    });
    return _nGram.filter(ite.expand((x) => x));
  }
}
class EnglishMinimalStemFilter implements TokenFilter {
  @override Iterable<String> filter(Iterable<String> tokens) {
    return tokens.map((token) => _stem(token));
  }
  String _stem(String text) {
    List<String> s = new Iterable.generate(text.length, (i) => text[i]).toList();
    return s.getRange(0, _doStem(s, s.length)).join();
  }
  int _doStem(List<String> s, int len) {
    if (len < 3 || s[len-1] != 's')
      return len;

    switch(s[len-2]) {
      case 'u':
      case 's': return len;
      case 'e':
        if (len > 3 && s[len-3] == 'i' && s[len-4] != 'a' && s[len-4] != 'e') {
          s[len - 3] = 'y';
          return len - 2;
        }
        if (s[len-3] == 'i' || s[len-3] == 'a' || s[len-3] == 'o' || s[len-3] == 'e')
          return len; /* intentional fallthrough */
        return len - 1;
      default: return len - 1;
    }
  }
}
class AliasFilter implements TokenFilter {
  final Map<String,List<String>> _dictionary;
  AliasFilter(this._dictionary);
  @override Iterable<String> filter(Iterable<String> tokens) {
    return tokens.map((token) {
      List<String> aliasList = _dictionary[token];
      if (aliasList != null) {
        return aliasList;
      } else {
        return [token];
      }
    }).expand((x) => x);
  }
}
class KanaFilter implements TokenFilter {
  @override Iterable<String> filter(Iterable<String> tokens) {
    return tokens.map((token) {
      String chars = '';
      for (int i = 0; i < token.length; i++) {
        String char = '';
        int code = token.codeUnitAt(i);
        if (StandardTokenizer.isKatakana(code)) {
          char = new String.fromCharCode(code - 96);
          char = char.replaceAll('゜', 'ー');
        } else {
          char = token[i];
        }
        chars = chars + char;
      }
      return chars;
    });
  }
}
class CharacterTypeFilter implements TokenFilter {
  final Set<int> _types;
  CharacterTypeFilter(this._types);
  @override Iterable<String> filter(Iterable<String> tokens) {
    return tokens.where((token) {
      if (token.length <= 0) return false;
      int code = token.codeUnitAt(0); // first character
      int charType = StandardTokenizer.unknown;
      if (StandardTokenizer.isDigit(code)) {
        charType = StandardTokenizer.digitId(code);
        return _types.contains(charType);
      }
      if (StandardTokenizer.isLatin(code)) {
        charType = StandardTokenizer.latinId(code);
        return _types.contains(charType);
      }
      if (StandardTokenizer.isIdeographic(code)) {
        charType = StandardTokenizer.ideographicId(code);
        return _types.contains(charType);
      }
      if (StandardTokenizer.isHiragana(code)) {
        charType = StandardTokenizer.hiraganaId(code);
        return _types.contains(charType);
      }
      if (StandardTokenizer.isKatakana(code)) {
        charType = StandardTokenizer.katakanaId(code);
        return _types.contains(charType);
      }
      return false;
    });
  }
}
class JapaneseCharacterTypeFilter extends CharacterTypeFilter {
  static final Set<int> _jTypes = new Set()
    ..add(StandardTokenizer.ideographic)
    ..add(StandardTokenizer.hiragana)
    ..add(StandardTokenizer.katakana);
  JapaneseCharacterTypeFilter() : super(_jTypes);
}
class TrimFilter implements TokenFilter {
  @override Iterable<String> filter(Iterable<String> tokens) {
    return tokens
        .map((e) => e.trim())
        .map((e) => _trimLeading(e))
        .where((String e) => e.length > 0);
  }
  String _trimLeading(String s) {
    if (s.length > 0) {
      String ret;
      if (StandardTokenizer.isControl(s.codeUnitAt(0))) {
        ret = s.substring(1);
      } else {
        ret = s;
      }
      return ret;
    }
    return s;
  }
}

class KeywordAnalyzer extends Analyzer {
  KeywordAnalyzer() {
    tokenizer = new KeywordTokenizer();
  }
}
class LowerCaseAnalyser extends Analyzer {
  LowerCaseAnalyser() {
    charFilters = [new IdeographicSpaceCharFilter()];
    tokenizer = new KeywordTokenizer();
    tokenFilters = [new LowerCaseFilter()];
  }
}
class WhitespaceAnalyzer extends Analyzer {
  WhitespaceAnalyzer() {
    tokenizer = new WhitespaceTokenizer();
    tokenFilters = [new LowerCaseFilter()];
  }
}
class RegExpAnalyzer extends Analyzer {
  final RegExp _regExp;
  RegExpAnalyzer(this._regExp) {
    tokenizer = new RegExpTokenizer(_regExp);
    tokenFilters = [new LowerCaseFilter()];
  }
}
class NGramAnalyzer extends Analyzer {
  final int _n;
  NGramAnalyzer(this._n) {
    tokenizer = new NGramTokenizer(_n);
    tokenFilters = [new LowerCaseFilter()];
  }
}
class EdgeNGramAnalyzer extends Analyzer {
  final int _n;
  EdgeNGramAnalyzer(this._n){
    tokenizer = new EdgeNGramTokenizer(_n);
    tokenFilters = [new LowerCaseFilter()];
  }
}
class ShingleAnalyzer extends Analyzer {
  List<CharFilter> _charFilters = [];
  final Tokenizer _tokenizer;
  final List<TokenFilter> _tokenFilters;
  final int _n;
  final String _separator;
  ShingleAnalyzer(this._charFilters, this._tokenizer, this._tokenFilters,
      this._n, this._separator) {
    charFilters = _charFilters;
    tokenizer = _tokenizer;
    List<TokenFilter> filters = [];
    filters.addAll(_tokenFilters);
    filters.add(new ShingleFilter(_n, _separator));
    tokenFilters = filters;
  }
}
class StandardAnalyzer extends Analyzer {
  List<String> _dictionary;
  int _max;
  StandardAnalyzer({List<String> dictionary, int max}) {
    _dictionary = (dictionary == null) ? [] : dictionary;
    _max = (max == null) ? 16 : max;
    charFilters = [new DigitSeparatorCharFilter()];
    tokenizer = new StandardTokenizer();
    tokenFilters
      = [new LowerCaseFilter(),
         new CompoundWordsFilter(_dictionary),
         new LengthFilter(1, _max),
         new EnglishStopWordsFilter(),
         new JapaneseStopWordsFilter(),
         new EnglishMinimalStemFilter()];
  }
}
class PhraseAnalyzer extends Analyzer {
  final Analyzer _analyzer;
  final String _separator;
  PhraseAnalyzer(this._analyzer, this._separator) {
    charFilters = _analyzer.charFilters;
    tokenizer = _analyzer.tokenizer;
    List<TokenFilter> filters = [];
    filters.addAll(_analyzer.tokenFilters);
    filters.add(new ShingleFilter(2, _separator));
    tokenFilters = filters;
  }
}
class TypoToleranceAnalyzer extends Analyzer {
  final Analyzer _analyzer;
  TypoToleranceAnalyzer(this._analyzer) {
    charFilters = _analyzer.charFilters;
    tokenizer = _analyzer.tokenizer;
    List<TokenFilter> filters = [];
    filters.addAll(_analyzer.tokenFilters);
    filters.add(new TypoToleranceFilter(4));
    tokenFilters = filters;
  }
}
class UniGramAnalyzer extends Analyzer {
  UniGramAnalyzer() {
    tokenizer = new NGramTokenizer(1);
  }
}
class BiGramAnalyzer extends Analyzer {
  BiGramAnalyzer() {
    tokenizer = new NGramTokenizer(2);
  }
}

class StandardTokenizer implements Tokenizer {
  static int control = 0;
  static int digit = 1;
  static int latin = 2;
  static int ideographic = 3;
  static int hiragana = 4;
  static int katakana = 5;
  static bool isControl(int code) {
    return controlId(code) != unknown;
  }
  static int controlId(int code) {
    return (0x0000 <= code && code <= 0x001F) ? control : unknown;
  }
  static bool isDigit(int code) {
    return digitId(code) != unknown;
  }
  static int digitId(int code) {
    return (0x0030 <= code && code <= 0x0039) ? digit : unknown;
  }
  static bool isLatin(int code) {
    return latinId(code) != unknown;
  }
  static int latinId(int code) {
    return ((0x0041 <= code && code <= 0x005A) || (0x0061 <= code && code <= 0x007A)) ? latin : unknown;
  }
  static bool _ideographic(int code) => (0x4E00 <= code && code <= 0x9FD5);
  static bool _ideographicA(int code) => (0x3400 <= code) && (code <= 0x4DFF);
  static bool _ideographicB(int code) => (0x20000 <= code) && (code <= 0x2A6FF);
  static bool _ideographicC(int code) => (0x2A700 <= code) && (code <= 0x2B734);
  static bool _ideographicD(int code) => (0x2B740 <= code) && (code <= 0x2B81F);
  static bool _ideographicE(int code) => (0x2B820 <= code) && (code <= 0x2CEAF);
  static bool isIdeographic(int code) {
    return ideographicId(code) != unknown;
  }
  static int ideographicId(int code) {
    if (_ideographic(code)
        || _ideographicA(code) || _ideographicB(code) || _ideographicC(code)
        || _ideographicD(code) || _ideographicE(code)) {
      return ideographic;
    }
    return unknown;
  }
  static bool isHiragana(int code) {
    return hiraganaId(code) != unknown;
  }
  static int hiraganaId(int code) {
    return (0x3041 <= code && code <= 0x3096) ?  hiragana : unknown;
  }
  static bool isKatakana(int code) {
    return katakanaId(code) != unknown;
  }
  static int katakanaId(int code) {
    return (0x30A1 <= code && code <= 0x30FA) || (code == 0x30FC) ? katakana : unknown;
  }
  int _charType(int code) {
    int a = digitId(code);
    if (a != unknown) return a;
    int b = latinId(code);
    if (b != unknown) return b;
    int c = ideographicId(code);
    if (c != unknown) return c;
    int d = hiraganaId(code);
    if (d != unknown) return d;
    int e = katakanaId(code);
    if (e != unknown) return e;
    return unknown;
  }
  static int unknown = -1;
  static int _initState = -10000;
  int _previousState = _initState;
  void _updateState(int type) {
    _previousState = type;
  }
  bool _shouldAdd(int type) => (type == unknown) ? false : true;
  bool _shouldBreakByType(int prevType, int type) => prevType != type;
  @override Iterable<String> tokenize(String text) {
    List<String> ret = [];
    List<String> tmpChars = [];
    for (int i = 0; i < text.length; i++) {
      int code = text.codeUnitAt(i);
      int type = _charType(code);
      int prevState = _previousState; _updateState(type);
      if (_shouldAdd(type)) {
        if (_shouldBreakByType(prevState, type)) _emit(tmpChars, ret);
        tmpChars.add(text[i]);
      } else {
        _emit(tmpChars, ret);
      }
    }
    _emit(tmpChars, ret);
    return ret;
  }
  void _emit(List tmpChars, List ret) {
    if (tmpChars.isNotEmpty) {
      String str = tmpChars.join();
      ret.add(str);
      tmpChars.clear();
    }
  }
}

class RomajiFilter implements TokenFilter {
  final Map<String,List<String>> _dictionary = {
  'あ':['a'],'い':['i','yi'],'う':['u','wu','whu'],'え':['e'],'お':['o'],
  'うぁ':['wha'],'うぃ':['whi','wi'],'うぇ':['whe','we'],'うぉ':['who'],
  'ゐ':['wi'],'ゑ':['we'],
  'ぁ':['la','xa'],'ぃ':['ili','xi','lyi','xyi'],'ぅ':['lu','xu'],'ぇ':['le','xe','lye','xye'],'ぉ':['lo','xo'],
  'か':['ka','ca'],'き':['ki'],'く':['ku','cu','qu'],'け':['ke'],'こ':['ko','co'],
  'きゃ':['kya'],'きぃ':['kyi'],'きゅ':['kyu'],'きぇ':['kye'],'きょ':['kyo'],
  'くゃ':['qya'],'くゅ':['qyu'],'くょ':['qyo'],
  'くぁ':['qwa','qa','kwa'],'くぃ':['qwi','qi','qyi'],'くぅ':['qwu'],'くぇ':['qwe','qe','qye'],'くぉ':['qwo','qo'],
  'が':['ga'],'ぎ':['gi'],'ぐ':['gu'],'げ':['ge'],'ご':['go'],
  'ぎゃ':['gya'],'ぎぃ':['gyi'],'ぎゅ':['gyu'],'ぎぇ':['gye'],'ぎょ':['gyo'],
  'ぐぁ':['gwa'],'ぐぃ':['gwi'],'ぐぅ':['gwu'],'ぐぇ':['gwe'],'ぐぉ':['gwo'],
  'ヵ':['lka','xka'],'ヶ':['lke','xke'],
  'さ':['sa'],'し':['si','ci','shi'],'す':['su'],'せ':['se','ce'],'そ':['so'],
  'しゃ':['sya','sha'],'しぃ':['syi'],'しゅ':['syu','shu'],'しぇ':['sye','she'],'しょ':['syo','sho'],
  'すぁ':['swa'],'すぃ':['swi'],'すぅ':['swu'],'すぇ':['swe'],'すぉ':['swo'],
  'ざ':['za'],'じ':['zi','ji'],'ず':['zu'],'ぜ':['ze'],'ぞ':['zo'],
  'じゃ':['zya','ja','jya'],'じぃ':['zyi','jyi'],'じゅ':['zyu','ju','jyu'],'じぇ':['zye','je','jye'],'じょ':['zyo','jo','jyo'],
  'た':['ta'],'ち':['ti','chi'],'つ':['tu','tsu'],'て':['te'],'と':['to'],
  'ちゃ':['tya','cha','cya'],'ちぃ':['tyi','cyi'],'ちゅ':['tyu','chu','cyu'],'ちぇ':['tye','che','cye'],'ちょ':['tyo','cho','cyo'],
  'つぁ':['tsa'],'つぃ':['tsi'],'つぇ':['tse'],'つぉ':['tso'],
  'てゃ':['tha'],'てぃ':['thi'],'てゅ':['thu'],'てぇ':['the'],'てょ':['tho'],
  'とぁ':['twa'],'とぃ':['twi'],'とぅ':['twu'],'とぇ':['twe'],'とぉ':['two'],
  'だ':['da'],'ぢ':['di'],'づ':['du'],'で':['de'],'ど':['do'],
  'ぢゃ':['dya'],'ぢぃ':['dyi'],'ぢゅ':['dyu'],'ぢぇ':['dye'],'ぢょ':['dyo'],
  'でゃ':['dha'],'でぃ':['dhi'],'でゅ':['dhu'],'でぇ':['dhe'],'でょ':['dho'],
  'どぁ':['dwa'],'どぃ':['dwi'],'どゅ':['dwu'],'どぇ':['dwe'],'どょ':['dwo'],
  'っ':['ltu','xtu','ltsu'],
  'っか':['kka','cca'],'っき':['kki'],'っく':['kku','ccu','qqu'],'っけ':['kke'],'っこ':['kko','cco'],
  'っさ':['ssa'],'っし':['ssi','cci'],'っす':['ssu'],'っせ':['sse','cce'],'っそ':['sso'],
  'った':['tta'],'っち':['tti'],'っつ':['ttu'],'って':['tte'],'っと':['tto'],
  'っぱ':['ppa'],'っぴ':['ppi'],'っぷ':['ppu'],'っぺ':['ppe'],'っぽ':['ppo'],
  'な':['na'],'に':['ni'],'ぬ':['nu'],'ね':['ne'],'の':['no'],
  'にゃ':['nya'],'にぃ':['nyi'],'にゅ':['nyu'],'にぇ':['nye'],'にょ':['nyo'],
  'は':['ha'],'ひ':['hi'],'ふ':['hu','fu'],'へ':['he'],'ほ':['ho'],
  'ひゃ':['hya'],'ひぃ':['hyi'],'ひゅ':['hyu'],'ひぇ':['hye'],'ひょ':['hyo'],
  'ふぁ':['fwa','fa'],'ふぃ':['fwi','fi','fyi'],'ふぅ':['fwu'],'ふぇ':['fwe','fe','fye'],'ふぉ':['fwo','fo'],
  'ふゃ':['fya'],'ふゅ':['fyu'],'ふょ':['fyo'],
  'ば':['ba'],'び':['bi'],'ぶ':['bu'],'べ':['be'],'ぼ':['bo'],
  'びゃ':['bya'],'びぃ':['byi'],'びゅ':['byu'],'びぇ':['bye'],'びょ':['byo'],
  'ヴぁ':['va'],'ヴぃ':['vi','vyi'],'ヴ':['vu'],'ヴぇ':['ve','vye'],'ヴぉ':['vo'],
  'ゔぁ':['va'],'ゔぃ':['vi','vyi'],'ゔ':['vu'],'ゔぇ':['ve','vye'],'ゔぉ':['vo'],
  'ヴゃ':['vya'],'ヴゅ':['vyu'],'ヴょ':['vyo'],
  'ゔゃ':['vya'],'ゔゅ':['vyu'],'ゔょ':['vyo'],
  'ぱ':['pa'],'ぴ':['pi'],'ぷ':['pu'],'ぺ':['pe'],'ぽ':['po'],
  'ぴゃ':['pya'],'ぴぃ':['pyi'],'ぴゅ':['pyu'],'ぴぇ':['pye'],'ぴょ':['pyo'],
  'ま':['ma'],'み':['mi'],'む':['mu'],'め':['me'],'も':['mo'],
  'みゃ':['mya'],'みぃ':['myi'],'みゅ':['myu'],'みぇ':['mye'],'みょ':['myo'],
  'や':['ya'],'ゆ':['yu'],'よ':['yo'],
  'ゃ':['lya','xya'],'ゅ':['lyu','xyu'],'ょ':['lyo','xyo'],
  'ら':['ra'],'り':['ri'],'る':['ru'],'れ':['re'],'ろ':['ro'],
  'りゃ':['rya'],'りぃ':['ryi'],'りゅ':['ryu'],'りぇ':['rye'],'りょ':['ryo'],
  'わ':['wa'],'を':['wo'],'ん':['n','nn',"n'",'xn'],
  'ゎ':['lwa','xwa'],
  'ー':['-']};
  @override Iterable<String> filter(Iterable<String> tokens) {
    return tokens.map((token) => _convert(token)).expand((x) => x)
        .where((String e) => e.length > 0);
  }
  List<String> _convert(String text) {
    if (text.length <= 0) return [text];
    if (!StandardTokenizer.isHiragana(text.codeUnitAt(0))) return [text];
    List<String> ret = [];
    bool skip = false;
    for (int i = 0; i < text.length; i++) {
      if (skip) {
        skip = false;
        continue;
      }
      int start = i;
      int end = i + 2;
      List<String> results = _lookup(text, start, end);
      if (results != null) {
        ret = _fill(ret, results);
        skip = true;
        continue;
      }
      end = i + 1;
      results = _lookup(text, start, end);
      if (results != null) {
        ret = _fill(ret, results);
        skip = false;
        continue;
      }
      ret = _fill(ret, [text[i]]);
    }
    return ret;
  }
  List<String> _fill(List<String> curList, List<String> results) {
    var ret = <String>[];
    if (curList.length == 0) {
      for (var r in results) {
        ret.add(r);
      }
    } else {
      for (var e in curList) {
        for (var r in results) {
          ret.add(e + r);
        }
      }
    }
    return ret;
  }
  List<String> _lookup(String text, int startIndex, int endIndex) {
    if (!(endIndex > text.length)) {
      String char = text.substring(startIndex, endIndex);
      return _doLookup(char);
    }
    return null;
  }
  List<String> _doLookup(String char) {
    List<String> ret = _dictionary[char];
    return (ret != null) ? ret : null;
  }
}
