// Copyright (c) 2016, Shinichiro Abe. All rights reserved. Use of this source code
// is governed by a BSD-style license that can be found in the LICENSE file.

import 'dart:async';
import 'dart:typed_data';

import '../../src/document/document.dart';
import '../../src/store/index_directory.dart';

class IndexReader {

  LiveDocsSource _liveDocsSource;
  IndexSource _indexSource;
  DataSource _dataSource;
  DocSet _liveDocs = new DocSet(new Set());
  Map<String,Map<String,DocSet>> _indexes = <String,Map<String,DocSet>>{};
  Map<String,Map<int,DocData>> _data = <String,Map<int,DocData>>{};

  IndexReader(IndexDirectory directory) {
    directory.initInput();
    _liveDocsSource = directory.inputLiveDocs();
    _indexSource = directory.inputIndex();
    _dataSource = directory.inputData();
  }

  Future read() async {
    if (_liveDocs.length == 0) {
      _liveDocs = await _liveDocsSource.read(_liveDocs);
    }
    if (_indexes.length == 0) {
      _indexes = await _indexSource.read(_indexes);
    }
    if (_data.length == 0) {
      _data = await _dataSource.read(_data);
    }
  }

  Future<DocSet> liveDocs() async {
    if (_liveDocs.length == 0) {
      _liveDocs = await _liveDocsSource.read(_liveDocs);
    }
    return _liveDocs;
  }

  Future<Iterable<String>> terms(String field) async {
    if (_indexes.length == 0 || _indexes[field] == null) {
      _indexes = await _indexSource.terms(field, _indexes);
    }
    Map<String,DocSet> r = _indexes[field];
    return (r == null) ? [] : r.keys;
  }

  Future<DocSet> postings(String field, String term) async {
    if (_indexes.length == 0 || _indexes[field] == null || _indexes[field][term] == null) {
      _indexes = await _indexSource.postings(field, term, _indexes);
    }
    DocSet r = _indexes[field][term];
    return (r == null) ? new DocSet(new Set()) : r;
  }

  Future<Iterable<String>> dataFields() async {
    if (_data.length == 0) {
      _data = await _dataSource.fields(_data);
    }
    Iterable<String> r = _data.keys;
    return (r == null) ? [] : r;
  }

  Future<List<String>> dataValues(String field, int docId) async {
    if (_data.length == 0 || _data[field] == null || _data[field][docId] == null) {
      _data = await _dataSource.values(field, docId, _data);
    }
    DocData r = _data[field][docId];
    return (r == null) ? [] : r.toList();
  }

  Future<int> docFreq(String field, String text) async {
    DocSet s = await postings(field, text);
    Set<int> r = s.toSet();
    return (r == null) ? -1 : r.length;
  }

  Future<int> maxDoc() async {
    List<String> r = (await terms('id')).toList();
    return (r == null) ? 0 : r.length;
  }

  Future<Document> document(int docId, {Iterable<String> fieldsToLoad}) async {
    var fields = (fieldsToLoad != null) ? fieldsToLoad : await dataFields();
    Document ret = new Document();
    for (String field in fields) {
      var text = await dataValues(field, docId);
      ret.add(new Field(field, text, null, null));
    }
    return ret;
  }

}

class DirectoryReader {
  static Future<IndexReader> open(IndexDirectory directory) async {
    IndexReader ir = new IndexReader(directory);
    await ir.read();
    return ir;
  }
}

class DocSet {
  Uint32List buffer;
  int length;
  DocSet(Set<int> docIds) {
    buffer = new Uint32List.fromList(docIds.toList());
    length = docIds.length;
  }
  void update(Set<int> docIds) {
    buffer = new Uint32List.fromList(docIds.toList());
    length = docIds.length;
  }
  Set<int> toSet() => buffer.toSet();
}

class DocData {
  Uint16List buffer;
  DocData(List<String> data) {
    buffer = new Uint16List.fromList(data.join('\$').codeUnits);
  }
  List<String> toList() => new String.fromCharCodes(buffer).split('\$');
}