// Copyright (c) 2016, Shinichiro Abe. All rights reserved. Use of this source code
// is governed by a BSD-style license that can be found in the LICENSE file.

import 'dart:async';
import 'dart:collection';

import '../../src/document/document.dart';
import '../../src/index/index_reader.dart';
import 'query.dart';

class IndexSearcher {

  final IndexReader _reader;

  IndexSearcher(this._reader);

  IndexReader get reader => _reader;

  Future<int> count(BoolQuery query) async {
    return (await search(query, await _reader.maxDoc(), scoreSort: ScoreSort.indexed())).totalHits;
  }

  Future<Document> doc(int docID, {Iterable<String> fieldsToLoad}) async {
    return await _reader.document(docID, fieldsToLoad: fieldsToLoad);
  }

  Future<TopDocs> search(BoolQuery query, int n, {ScoreSort scoreSort}) async {
    Set<int> filtered = await _getDocSet(Op.and, query.filters);
    if (query.filters.length > 0 && filtered.isEmpty) {
      return new TopDocs(0, [new ScoreDoc(-1, double.NAN)], filtered);
    }
    Set<int> curDocSet = await _getDocSet(query.op, query.termQueries);
    if (!filtered.isEmpty) {
      curDocSet = filtered.intersection(curDocSet);
    }
    if (curDocSet.isEmpty) {
      return new TopDocs(0, [new ScoreDoc(-1, double.NAN)], curDocSet);
    }

    scoreSort = (scoreSort == null) ? new MatchScoreSort() : scoreSort;
    Map<int,double> docIdScore = await scoreSort.scoreThenSort(query, curDocSet, _reader);

    int totalHits = docIdScore.length;
    int max = (totalHits < n) ? totalHits : n;
    List<ScoreDoc> scoreDocs = [];
    for (int docId in docIdScore.keys.toList().getRange(0, max)) {
      ScoreDoc scoreDoc = new ScoreDoc(docId, docIdScore[docId]);
      scoreDocs.add(scoreDoc);
    }
    return new TopDocs(totalHits, scoreDocs, curDocSet);
  }

  Future<Set<int>> _getDocSet(Op op, List<Query> queries) async {
    Set<int> ret = new Set();
    bool isFirst = true;
    for (Query query in queries) {
      Set<int> docIds = await query.match(this);
      if (docIds == null) docIds = new Set();
      if (isFirst) {
        ret = docIds;
        isFirst = false;
        continue;
      }
      ret = unionDocSet(op, ret, docIds);
    }
    return ret;
  }

  static Set<int> unionDocSet(Op op, Set<int> from, Set<int> other) {
    Set<int> ret = new Set.from(from);
    switch (op) {
      case Op.and:
        ret = ret.intersection(other);
        break;
      case Op.or:
        ret = ret.union(other);
        break;
    }
    return ret;
  }

}

class TopDocs {
  final int totalHits;
  final List<ScoreDoc> scoreDocs;
  final Set<int> docSet;
  TopDocs(this.totalHits, this.scoreDocs, this.docSet);
}

class ScoreDoc {
  final int doc;
  final double score;
  ScoreDoc(this.doc, this.score);
}

abstract class ScoreSort {
  static ScoreSort indexed() => new DocIdSort(false);
  static ScoreSort reversed() => new DocIdSort(true);
  Future<Map<int,double>> scoreThenSort(BoolQuery query, Set<int> docSet, IndexReader reader);
}

class DocIdSort implements ScoreSort {
  final bool _reversed;
  DocIdSort(this._reversed);
  Future<Map<int,double>> scoreThenSort(BoolQuery query, Set<int> docSet, IndexReader reader) async {
    var docList = docSet.toList();
    if (!_reversed) {
      docList.sort();
      return new Map.fromIterable(docList, key: (int e) => e, value: (int e) => 1.00);
    } else {
      return new Map.fromIterable(docList.reversed.toList(), key: (int e) => e, value: (int e) => 1.00);
    }
  }
}

class MatchScoreSort implements ScoreSort {
  Future<Map<int,double>> scoreThenSort(BoolQuery query, Set<int> docSet, IndexReader reader) async {
    Map<int,double> m = {};
    var maxDoc = await reader.maxDoc();
    for (Query q in query.termQueries) {
      for (int docId in q.matchedDocIds) {
        if (docSet.contains(docId)) {
          if (m[docId] == null) {
            m[docId] = q.score() + (maxDoc/(docId+1) * 0.00000001);
          } else {
            m[docId] = m[docId] + q.score();
          }
        }
      }
    }
    SplayTreeMap<double,int> sort
      = new SplayTreeMap.fromIterables(m.values, m.keys, (a, b) {
        return b.compareTo(a);
      });
    return new Map.fromIterables(sort.values, sort.keys);
  }
}