// Copyright (c) 2016, Shinichiro Abe. All rights reserved. Use of this source code
// is governed by a BSD-style license that can be found in the LICENSE file.

import 'dart:async';
import 'dart:convert';
import 'dart:io';

import '../../src/index/index_reader.dart';

abstract class IndexDirectory {
  Future initOutput();
  Future outputLiveDocs(Set<int> docIds);
  Future outputIndex(String name, Map<String,DocSet> index);
  Future outputData(String name, Map<int,DocData> data);
  void initInput();
  LiveDocsSource inputLiveDocs();
  IndexSource inputIndex();
  DataSource inputData();
}
// ignore: one_member_abstracts
abstract class LiveDocsStreams {
  Future readFully(DocSet to);
}
abstract class IndexStreams {
  Future readFully(Map<String,Map<String,DocSet>> to);
  Future readTerms(String field, Map<String,Map<String,DocSet>> to);
  Future readPostings(String field, String term, Map<String,Map<String,DocSet>> to);
}
abstract class DataStreams {
  Future readFully(Map<String,Map<int,DocData>> to);
  Future readFields(Map<String,Map<int,DocData>> to);
  Future readValues(String field, int docId, Map<String,Map<int,DocData>> to);
}

class LiveDocsSource {
  final LiveDocsStreams _streams;
  LiveDocsSource(this._streams);
  Future<DocSet> read(DocSet to) async {
    await _streams.readFully(to);
    return to;
  }
}
class IndexSource {
  final IndexStreams _streams;
  IndexSource(this._streams);
  Future<Map<String,Map<String,DocSet>>> read(Map<String,Map<String,DocSet>> to) async {
    await _streams.readFully(to);
    return to;
  }
  Future<Map<String,Map<String,DocSet>>> terms(String field, Map<String,Map<String,DocSet>> to) async {
    await _streams.readTerms(field, to);
    return to;
  }
  Future<Map<String,Map<String,DocSet>>> postings(String field, String term, Map<String,Map<String,DocSet>> to) async {
    await _streams.readPostings(field, term, to);
    return to;
  }
}
class DataSource {
  DataStreams _streams;
  DataSource(this._streams);
  Future<Map<String,Map<int,DocData>>> read(Map<String,Map<int,DocData>> to) async {
    await _streams.readFully(to);
    return to;
  }
  Future<Map<String,Map<int,DocData>>> fields(Map<String,Map<int,DocData>> to) async {
    await _streams.readFields(to);
    return to;
  }
  Future<Map<String,Map<int,DocData>>> values(String field, int docId, Map<String,Map<int,DocData>> to) async {
    await _streams.readValues(field, docId, to);
    return to;
  }
}

class CompositeIndexDirectory implements IndexDirectory {
  final List<IndexDirectory> _directories;
  CompositeIndexDirectory(this._directories);
  @override Future initOutput() { return new Future.value(Null); }
  @override Future outputLiveDocs(Set<int> docIds) {
    throw new Exception('UnsupportedOperationException');
  }
  @override Future outputIndex(String name, Map<String,DocSet> index) {
    throw new Exception('UnsupportedOperationException');
  }
  @override Future outputData(String name, Map<int,DocData> data) {
    throw new Exception('UnsupportedOperationException');
  }
  @override void initInput() { }
  @override LiveDocsSource inputLiveDocs() => new LiveDocsSource(new CompositeLiveDocsStreams(_directories));
  @override IndexSource inputIndex() => new IndexSource(new CompositeIndexStreams(_directories));
  @override DataSource inputData() => new DataSource(new CompositeDataStreams(_directories));
}
class CompositeLiveDocsStreams extends LiveDocsStreams {
  final List<IndexDirectory> _directories;
  CompositeLiveDocsStreams(this._directories);
  @override Future readFully(DocSet to) async {
    Set<int> ret = new Set();
    bool isFirst = true;
    int prevMaxDoc = 0;
    for (IndexDirectory directory in _directories) {
      Map<String,Map<String,DocSet>> indexes = {};
      indexes = await directory.inputIndex().read(indexes);
      int maxDoc = indexes['id'].keys.length;
      Set<int> docIds = (await directory.inputLiveDocs().read(to)).toSet();
      if (isFirst) {
        ret = docIds;
        isFirst = false;
        prevMaxDoc = maxDoc;
        continue;
      }
      ret.addAll(docIds.map((docId) => docId + prevMaxDoc).toSet());
      prevMaxDoc = maxDoc;
    }
    to.update(ret);
  }
}
class CompositeIndexStreams extends IndexStreams {
  final List<IndexDirectory> _directories;
  CompositeIndexStreams(this._directories);
  bool _readFully = false;
  @override Future readFully(Map<String,Map<String,DocSet>> to) async {
    _readFully = true;
    Map<String,Map<String,DocSet>> ret = {};
    int prevMaxDoc = 0;
    for (IndexDirectory directory in _directories) {
      Map<String,Map<String,DocSet>> indexes = {};
      indexes = await directory.inputIndex().read(indexes);
      for (String field in indexes.keys) {
        if (ret[field] == null) {
          ret[field] = indexes[field];
          continue;
        }
        Map<String,DocSet> index = indexes[field];
        for (String term in index.keys) {
          Set<int> docIds = index[term].toSet();
          Set<int> newDocIds = docIds.map((id) => id + prevMaxDoc).toSet();
          if (ret[field][term] == null) {
            ret[field][term] = new DocSet(newDocIds);
          } else {
            Set<int> cur = ret[field][term].toSet();
            cur.addAll(newDocIds);
            ret[field][term] = new DocSet(cur);
          }
        }
      }
      int maxDoc = indexes['id'].keys.length;
      prevMaxDoc = maxDoc;
    }
    for (String field in ret.keys) {
      to[field] = ret[field];
    }
  }
  @override Future readTerms(String field, Map<String,Map<String,DocSet>> to) async {
    if (!_readFully) await readFully(to);
  }
  @override Future readPostings(String field, String term, Map<String,Map<String,DocSet>> to) async {
    if (!_readFully) await readFully(to);
  }
}
class CompositeDataStreams extends DataStreams {
  final List<IndexDirectory> _directories;
  CompositeDataStreams(this._directories);
  bool _readFully = false;
  @override Future readFully(Map<String,Map<int,DocData>> to) async {
    _readFully = true;
    Map<String,Map<int,DocData>> ret = {};
    int prevMaxDoc = 0;
    for (IndexDirectory directory in _directories) {
      Map<String,Map<int,DocData>> data = {};
      data = await directory.inputData().read(data);
      for (String field in data.keys) {
        if (ret[field] == null) {
          ret[field] = data[field];
          continue;
        }
        Map<int,DocData> docs = data[field];
        for (int docId in docs.keys) {
          int newDocIds = docId + prevMaxDoc;
          ret[field][newDocIds] = docs[docId];
        }
      }
      int maxDoc = data['id'].keys.length;
      prevMaxDoc = maxDoc;
    }
    for (String field in ret.keys) {
      to[field] = ret[field];
    }
  }
  @override Future readFields(Map<String,Map<int,DocData>> to) async {
    if (!_readFully) await readFully(to);
  }
  @override Future readValues(String field, int docId, Map<String,Map<int,DocData>> to) async {
    if (!_readFully) await readFully(to);
  }
}

class RAMIndexDirectory implements IndexDirectory {

  DocSet _liveDocs = new DocSet(new Set());
  Map<String,Map<String,DocSet>> _ramIndex = {};
  Map<String,Map<int,DocData>> _ramData = {};

  @override Future initOutput() { return new Future.value(Null); }

  @override Future outputLiveDocs(Set<int> docIds) {
    _liveDocs = new DocSet(docIds);
    return new Future.value(Null);
  }

  @override Future outputIndex(String name, Map<String,DocSet> index) {
    _ramIndex[name] = index;
    return new Future.value(Null);
  }

  @override Future outputData(String name, Map<int,DocData> data) {
    _ramData[name] = data;
    return new Future.value(Null);
  }

  @override void initInput() { }
  @override LiveDocsSource inputLiveDocs() => new LiveDocsSource(new RAMLiveDocsStreams(_liveDocs));
  @override IndexSource inputIndex() => new IndexSource(new RAMIndexStreams(_ramIndex));
  @override DataSource inputData() => new DataSource(new RAMDataStreams(_ramData));

}
class RAMLiveDocsStreams extends LiveDocsStreams {
  final DocSet _liveDocs;
  RAMLiveDocsStreams(this._liveDocs);
  @override Future readFully(DocSet to) async {
    to.update(_liveDocs.toSet());
  }
}
class RAMIndexStreams extends IndexStreams {
  final Map<String,Map<String,DocSet>> _ramIndex;
  RAMIndexStreams(this._ramIndex);
  bool _readFully = false;
  @override Future readFully(Map to) async {
    for (String field in _ramIndex.keys) {
      to[field] = _ramIndex[field];
    }
    _readFully = true;
  }
  @override Future readTerms(String field, Map to) async {
    if (!_readFully) await readFully(to);
  }
  @override Future readPostings(String field, String term, Map to) async {
    if (!_readFully) await readFully(to);
  }
}
class RAMDataStreams extends DataStreams {
  final Map<String,Map<int,DocData>> _ramData;
  RAMDataStreams(this._ramData);
  bool _readFully = false;
  @override Future readFully(Map to) async {
    for (String field in _ramData.keys) {
      to[field] = _ramData[field];
    }
    _readFully = true;
  }
  @override Future readFields(Map to) async {
    if (!_readFully) await readFully(to);
  }
  @override Future readValues(String field, int docId, Map to) async {
    if (!_readFully) await readFully(to);
  }
}

class FSIndexDirectory implements IndexDirectory {

  final Directory _dir;

  FSIndexDirectory(this._dir);

  @override Future initOutput() {
    if (_dir.existsSync()) _dir.deleteSync(recursive: true);
    if (!_dir.existsSync()) _dir.createSync(recursive: true);
    return new Future.value(Null);
  }

  @override Future outputLiveDocs(Set<int> docIds) {
    File liveDocsFile = new File('${_dir.path}/.liv');
    if (liveDocsFile.existsSync()) liveDocsFile.deleteSync(); //mutable
    if (!liveDocsFile.existsSync()) liveDocsFile.createSync();
    liveDocsFile.writeAsStringSync(docIds.join(','), mode: FileMode.APPEND, encoding: UTF8);
    return new Future.value(Null);
  }

  @override Future outputIndex(String name, Map<String,DocSet> index) {
    File indexFile = new File('${_dir.path}/$name.idx');
    if (!indexFile.existsSync()) indexFile.createSync();
    for (String key in index.keys) {
      String value = index[key].toSet().join(',');
      String line = '$key=>$value';
      indexFile.writeAsStringSync(line+'\n', mode: FileMode.APPEND, encoding: UTF8);
    }
    return new Future.value(Null);
  }

  @override Future outputData(String name, Map<int,DocData> data) {
    File storeFile = new File('${_dir.path}/$name.data');
    if (!storeFile.existsSync()) storeFile.createSync();
    for (int key in data.keys) {
      String value = data[key].toList().join('\$');
      String line = '${key.toString()}=>$value';
      storeFile.writeAsStringSync(line+'\n', mode: FileMode.APPEND, encoding: UTF8);
    }
    return new Future.value(Null);
  }

  @override void initInput() { }
  @override LiveDocsSource inputLiveDocs() => new LiveDocsSource(new FSLiveDocsStreams(_dir));
  @override IndexSource inputIndex() => new IndexSource(new FSIndexStreams(_dir));
  @override DataSource inputData() => new DataSource(new FSDataStreams(_dir));

}
class FSLiveDocsStreams extends LiveDocsStreams {
  final Directory _dir;
  FSLiveDocsStreams(this._dir);
  @override Future readFully(DocSet to) async {
    for (String path in FSIndexUtils.getFileList(_dir, '.liv')) {
      File indexFile = new File(path);
      List<String> lines = indexFile.readAsLinesSync();
      String s = lines[0];
      Set<int> docIds = s.split(',').map((s) => int.parse(s)).toSet();
      to.update(docIds);
      break;
    }
  }
}
class FSIndexStreams extends IndexStreams {
  final Directory _dir;
  FSIndexStreams(this._dir);
  bool _readFully = false;
  @override Future readFully(Map to) async {
    for (String path in FSIndexUtils.getFileList(_dir, '.idx')) {
      String name = FSIndexUtils.getName(path, '.idx');
      File indexFile = new File(path);
      List<String> lines = indexFile.readAsLinesSync();
      var invertedIndex = <String,DocSet>{};
      for (String line in lines) {
        List<String> kv = line.split('=>');
        String key = kv[0];
        Set<int> values = kv[1].split(',').map((s) => int.parse(s)).toSet();
        invertedIndex[key] = new DocSet(values);
      }
      to[name] = invertedIndex;
    }
    _readFully = true;
  }
  @override Future readTerms(String field, Map to) async {
    if (!_readFully) await readFully(to);
  }
  @override Future readPostings(String field, String term, Map to) async {
    if (!_readFully) await readFully(to);
  }
}
class FSDataStreams extends DataStreams {
  final Directory _dir;
  FSDataStreams(this._dir);
  bool _readFully = false;
  @override Future readFully(Map to) async {
    for (String path in FSIndexUtils.getFileList(_dir, '.data')) {
      String name = FSIndexUtils.getName(path, '.data');
      File dataFile = new File(path);
      String str = dataFile.readAsStringSync();
      List<String> lines = str.split('\n');
      var storedData = <int,DocData>{};
      for (String line in lines) {
        List<String> kv = line.split('=>');
        if (kv.length == 2) {
          int key = int.parse(kv[0]);
          List<String> values = kv[1].split('\$');
          storedData[key] = new DocData(values);
        }
      }
      to[name] = storedData;
    }
    _readFully = true;
  }
  @override Future readFields(Map to) async {
    if (!_readFully) await readFully(to);
  }
  @override Future readValues(String field, int docId, Map to) async {
    if (!_readFully) await readFully(to);
  }
}
class FSIndexUtils {
  static Iterable<String> getFileList(Directory dir, String extension) =>
      dir.listSync(recursive: false, followLinks: false)
          .map((e) => e.path)
          .where((path) => path.endsWith(extension));

  static String getName(String path, String extension) =>
      path.split('/').last.replaceAll(extension, '');
}

class FSIndexJsonConverter {
  final Directory _dir;
  FSIndexDirectory _fsDir;

  FSIndexJsonConverter(this._dir) {
    _fsDir = new FSIndexDirectory(_dir);
  }

  Future run() async {
    Map ret = {};

    Map<String,Map<String,DocSet>> inputIndex = {};
    inputIndex = await _fsDir.inputIndex().read(inputIndex);
    Map<String,Map<String,Map<String,String>>> idx = {};
    for (String fieldName in inputIndex.keys) {
      Map<String,DocSet> fieldValue = inputIndex[fieldName];
      Map<String,Map<String,String>> terms = {};
      int i = 0;
      for (String term in fieldValue.keys) {
        Map<String,String> termInfo = {};
        Set<int> postings = fieldValue[term].toSet();
        termInfo["postings"] = postings.join(",");
        termInfo["term"] = term;
        terms[i.toString()] = termInfo;
        i++;
      }
      idx[fieldName] = terms;
    }

    Map<String,Map<int,DocData>> inputData = {};
    inputData = await _fsDir.inputData().read(inputData);
    Map<String,Map<String,String>> data = {};
    for (String fieldName in inputData.keys) {
      Map<int,DocData> fieldValue = inputData[fieldName];
      Map<String,String> fieldData = {};
      for (int docId in fieldValue.keys) {
        List<String> data = fieldValue[docId].toList();
        fieldData[docId.toString()] = data.join("\$");
      }
      data[fieldName] = fieldData;
    }

    ret["idx"] = idx;
    ret["data"] = data;

    _outJson(ret, "fsindex.json");
  }

  void _outJson(Map obj, String outFilePath) {
    File outFile = new File(outFilePath);
    if (outFile.existsSync()) outFile.deleteSync();
    if (!outFile.existsSync()) outFile.createSync();
    outFile.writeAsStringSync(JSON.encode(obj), mode: FileMode.APPEND, encoding: UTF8);
  }

}