// Copyright (c) 2016, Shinichiro Abe. All rights reserved. Use of this source code
// is governed by a BSD-style license that can be found in the LICENSE file.

import 'dart:async';

import '../../src/document/document.dart';
import '../../src/index/index_reader.dart';
import '../../src/search/index_searcher.dart';
import '../../src/search/query.dart';
import 'index_directory.dart';

class IndexWriter {

  final IndexDirectory _directory;

  IndexWriter(this._directory);

  Future write(Iterable<Document> documents) async {
    int docId = 0;
    Set<int> liveDocs = new Set();
    Map<String,InvertedIndexBuilder> indexes = {};
    Map<String,StoredDataBuilder> data = {};
    for (Document doc in documents) {
      _buildIndex(docId, doc, liveDocs, indexes, data);
      docId++;
    }
    await _directory.initOutput();
    await _storeLiveDocs(liveDocs);
    await _storeIndexes(indexes);
    await _storeData(data);
  }

  void _buildIndex(int docId, Document document,
      Set<int> liveDocs,
      Map<String,InvertedIndexBuilder> indexes,
      Map<String,StoredDataBuilder> data) {
    liveDocs.add(docId);
    for (Field field in document.fields) {
      String name = field.name;
      List<String> text = field.text;
      var tokensIte = new Iterable.generate(text.length,
            (i) => field.analyzer.getTokens(text[i]));
      Iterable<String> tokens = tokensIte.expand((x) => x);
      if (!indexes.containsKey(name)) {
        indexes[name] = new InvertedIndexBuilder().append(tokens, docId);
      } else {
        indexes[name].append(tokens, docId);
      }
      if (field.stored) {
        if (!data.containsKey(name)) {
          data[name] = new StoredDataBuilder().append(docId, text);
        } else {
          data[name].append(docId, field.text);
        }
      }
    }
  }

  Future _storeLiveDocs(Set<int> liveDocs) async {
    await _directory.outputLiveDocs(liveDocs);
  }

  Future _storeIndexes(Map<String,InvertedIndexBuilder> indexes) async {
    for (String name in indexes.keys) {
      await _directory.outputIndex(name, indexes[name].build());
    }
  }

  Future _storeData(Map<String,StoredDataBuilder> data) async {
    for (String name in data.keys) {
      await _directory.outputData(name, data[name].build());
    }
  }

  Future delete(BoolQuery query) async {
    IndexReader reader = await DirectoryReader.open(_directory);
    IndexSearcher searcher = new IndexSearcher(reader);
    TopDocs hits = await searcher.search(query, await reader.maxDoc(),
        scoreSort: ScoreSort.indexed());
    Set<int> docIds = hits.scoreDocs.map((scoreDoc) => scoreDoc.doc).toSet();
    DocSet liveDocs = await reader.liveDocs();
    Set<int> newLiveDocs = liveDocs.toSet().difference(docIds);
    await _storeLiveDocs(newLiveDocs);
  }

  static Future copyFrom(CompositeIndexDirectory from, IndexDirectory to) async {
    await to.initOutput();
    DocSet liveDocs = new DocSet(new Set());
    liveDocs = await from.inputLiveDocs().read(liveDocs);
    await to.outputLiveDocs(liveDocs.toSet());
    Map<String,Map<String,DocSet>> indexes = {};
    indexes = await from.inputIndex().read(indexes);
    for (String field in indexes.keys) {
      await to.outputIndex(field, indexes[field]);
    }
    Map<String,Map<int,DocData>> data = {};
    data = await from.inputData().read(data);
    for (String field in data.keys) {
      await to.outputData(field, data[field]);
    }
  }
}

class InvertedIndexBuilder {

  var _invertedIndex = <String,Set<int>>{};

  InvertedIndexBuilder append(Iterable<String> tokens, int docId) {
    for (String token in tokens) {
      if(!_invertedIndex.containsKey(token)) {
        Set<int> set = new Set()..add(docId);
        _invertedIndex[token] = set;
      } else {
        Set<int> set = _invertedIndex[token];
        if (!set.contains(docId)) {
          set.add(docId);
        }
      }
    }
    return this;
  }

  Map<String,DocSet> build() {
    var ret = <String,DocSet>{};
    List<String> terms = _invertedIndex.keys.toList();
    terms.sort();
    for (String term in terms) {
      ret[term] = new DocSet(_invertedIndex[term]);
    }
    return ret;
  }
}

class StoredDataBuilder {

  var _storedData = <int,DocData>{};

  StoredDataBuilder append(int docId, List<String> text) {
    _storedData[docId] = new DocData(text);
    return this;
  }

  Map<int,DocData> build() {
    return _storedData;
  }

}