// Copyright (c) 2016, Shinichiro Abe. All rights reserved. Use of this source code
// is governed by a BSD-style license that can be found in the LICENSE file.

import 'dart:async';

import 'package:sembast/sembast.dart';
import 'package:sembast/sembast_io.dart';

import '../../../src/index/index_reader.dart';
import '../../../src/store/index_directory.dart';

class SembastIndexDirectory implements IndexDirectory {
  final DatabaseFactory _dbFactory = new IoDatabaseFactory();
  final List<String> _indexedFields;
  final List<String> _storedFields;
  final String _dbPath;
  SembastIndexDirectory(this._dbPath,this._indexedFields,this._storedFields);

  @override Future initOutput() async {
    for (String field in _indexedFields) {
      await _dbFactory.deleteDatabase('$_dbPath/$field.idx');
    }
    for (String field in _storedFields) {
      await _dbFactory.deleteDatabase('$_dbPath/$field.data');
    }
  }

  @override Future outputLiveDocs(Set<int> docIds) async {
    Database db = await _dbFactory.openDatabase('$_dbPath/.liv');
    await db.put(docIds.join(','), '__LIVEDOCS__');
    db.close();
  }

  @override Future outputIndex(String name, Map<String,DocSet> index) async {
    Database db = await _dbFactory.openDatabase('$_dbPath/$name.idx');
    List<String> terms = [];
    for (String term in index.keys) {
      terms.add(term);
      Set<int> postings = index[term].toSet();
      await db.put(postings.join(","), term);
    }
    await db.put(terms.join('\$'), '__TERMS__');
    db.close();
  }

  @override Future outputData(String name, Map<int,DocData> data) async {
    Database db = await _dbFactory.openDatabase('$_dbPath/$name.data');
    List<int> docIds = [];
    for (int docId in data.keys) {
      docIds.add(docId);
      List<String> datum = data[docId].toList();
      await db.put(datum.join("\$"), docId);
    }
    await db.put(docIds.join('\$'), '__DOCIDS__');
    db.close();
  }

  @override void initInput() { }

  @override LiveDocsSource inputLiveDocs() {
    return new LiveDocsSource(new SembastLiveDocsStreams(_dbPath,_dbFactory));
  }

  @override IndexSource inputIndex() {
    _indexStreams = new SembastIndexStreams(_indexedFields,_dbPath,_dbFactory);
    return new IndexSource(_indexStreams);
  }

  @override DataSource inputData() {
    _dataStreams = new SembastDataStreams(_storedFields,_dbPath,_dbFactory);
    return new DataSource(_dataStreams);
  }
  SembastIndexStreams _indexStreams;
  SembastDataStreams _dataStreams;
  bool loaded() => _indexStreams.loaded && _dataStreams.loaded;
}
class SembastLiveDocsStreams extends LiveDocsStreams {
  final DatabaseFactory _dbFactory;
  final String _dbPath;
  SembastLiveDocsStreams(this._dbPath,this._dbFactory);
  @override Future readFully(DocSet to) async {
    Database db = await _dbFactory.openDatabase('$_dbPath/.liv');
    String s = await db.get('__LIVEDOCS__');
    Set<int> docIds = s.split(',').map((s) => int.parse(s)).toSet();
    to.update(docIds);
  }
}
class SembastIndexStreams extends IndexStreams {
  final List<String> _indexedFields;
  final String _dbPath;
  final DatabaseFactory _dbFactory;
  SembastIndexStreams(this._indexedFields,this._dbPath,this._dbFactory);
  bool _readFully;
  bool loaded = false;

  @override Future readFully(Map to) async {
    _readFully = true;
    for (String field in _indexedFields) {
      Database db = await _dbFactory.openDatabase('$_dbPath/$field.idx');
      String s = await db.get('__TERMS__');
      List<String> terms  = s.split('\$').toList();
      Map<String, DocSet> invertedIndex = {};
      for (String term in terms) {
        String ss = await db.get(term);
        String key = term;
        Set<int> values = ss.split(',').map((s) => int.parse(s)).toSet();
        invertedIndex[key] = new DocSet(values);
      }
      to[field] = invertedIndex;
    }
    loaded = true;
  }

  @override Future readTerms(String field, Map<String,Map<String,DocSet>> to) async {
    if (!_readFully) await readFully(to);
  }

  @override Future readPostings(String field, String term, Map<String,Map<String,DocSet>> to) async {
    if (!_readFully) await readFully(to);
  }
}
class SembastDataStreams extends DataStreams {
  final List<String> _storedFields;
  final String _dbPath;
  final DatabaseFactory _dbFactory;
  SembastDataStreams(this._storedFields,this._dbPath,this._dbFactory);
  bool _readFully;
  bool loaded = false;

  @override Future readFully(Map to) async {
    _readFully = true;
    for (String field in _storedFields) {
      Database db = await _dbFactory.openDatabase('$_dbPath/$field.data');
      String s = await db.get('__DOCIDS__');
      List<int> docIds = s.split('\$').map((s) => int.parse(s)).toList();
      Map<int,DocData> storedData = {};
      for (int docId in docIds) {
        List<String> values = s.split('\$').toList();
        storedData[docId] = new DocData(values);
      }
      to[field] = storedData;
    }
    loaded = true;
  }

  @override Future readFields(Map<String,Map<int,DocData>> to) async {
    if (!_readFully) await readFully(to);
  }

  @override Future readValues(String field, int docId, Map<String,Map<int,DocData>> to) async {
    if (!_readFully) await readFully(to);
  }
}