// Copyright (c) 2016, Shinichiro Abe. All rights reserved. Use of this source code
// is governed by a BSD-style license that can be found in the LICENSE file.

import 'dart:io';

import 'package:ducene/document.dart';
import 'package:ducene/index.dart';
import 'package:ducene/search.dart';
import 'package:ducene/store.dart';
import 'package:test/test.dart';

void main() {
  group('A group of writer', () {
    Directory dir1 = new Directory('test-dir-writer1');
    IndexDirectory d1;
    Directory dir2 = new Directory('test-dir-writer2');
    IndexDirectory d2;
    Directory dir3 = new Directory('test-dir-writer3');
    IndexDirectory d3;

    setUp(() {
      if (dir1.existsSync()) dir1.deleteSync(recursive: true);
      d1 = new FSIndexDirectory(dir1);
      if (dir2.existsSync()) dir2.deleteSync(recursive: true);
      d2 = new FSIndexDirectory(dir2);
      if (dir3.existsSync()) dir3.deleteSync(recursive: true);
      d3 = new FSIndexDirectory(dir3);
    });

    tearDown(() {
      if (dir1.existsSync()) dir1.deleteSync(recursive: true);
      if (dir2.existsSync()) dir2.deleteSync(recursive: true);
      if (dir3.existsSync()) dir3.deleteSync(recursive: true);
    });

    test('test rewrite', () async {
      IndexDirectory indexDirectory = new RAMIndexDirectory();
      IndexWriter writer = new IndexWriter(indexDirectory);
      List<Document> docs = [];
      for (int i = 0; i < 10; i++) {
        Document doc = new Document();
        doc.append("id", i.toString());
        docs.add(doc);
      }
      await writer.write(docs);

      expect(await (await DirectoryReader.open(indexDirectory)).maxDoc(), 10);

      docs = [];
      for (int i = 0; i < 3; i++) {
        Document doc = new Document();
        doc.append("id", i.toString());
        docs.add(doc);
      }
      await writer.write(docs);

      expect(await (await DirectoryReader.open(indexDirectory)).maxDoc(), 3);
    });

    test('test index and search get()', () async {
      IndexDirectory indexDirectory = new RAMIndexDirectory();
      IndexWriter writer = new IndexWriter(indexDirectory);
      Document doc = new Document();
      doc.append("id", '1');
      doc.append("text1", 'text1');
      doc.append("text2", ['text2-1', 'text2-2']);
      doc.append("text3", 'text3', stored: false);
      doc.append("text4", ['text4-1', 'text4-2'], stored: false);
      await writer.write([doc]);

      IndexReader reader = await DirectoryReader.open(indexDirectory);
      IndexSearcher searcher = new IndexSearcher(reader);
      TopDocs docs = await searcher.search(new BoolQuery().append("id", '1'), 10, scoreSort: ScoreSort.indexed());
      expect(docs.totalHits, 1);
      ScoreDoc scoreDoc = docs.scoreDocs[0];
      Document d = await searcher.doc(scoreDoc.doc);
      expect(d.get("id"), '1');
      expect(d.get("text1"), 'text1');
      expect(d.getValues("text2"), ['text2-1', 'text2-2']);
      expect(d.get("text3"), isNull);
      expect(d.getValues("text4"), isNull);
    });

    test('test delete simple', () async {
      IndexWriter writer = new IndexWriter(d1);
      await writer.write([
        new Document().append("id", '1').append("all","x"),
        new Document().append("id", '2').append("all","x"),
        new Document().append("id", '3').append("all","x")]);

      IndexReader reader = await DirectoryReader.open(d1);
      IndexSearcher searcher = new IndexSearcher(reader);
      expect(await searcher.count(new BoolQuery().append("all", "x")), 3);

      await writer.delete(new BoolQuery().append("id", "2"));

      reader = await DirectoryReader.open(d1);
      searcher = new IndexSearcher(reader);
      expect(await searcher.count(new BoolQuery().append("all", "x")), 2);
    });

    test('test copyFrom', () async {
      IndexWriter writer = new IndexWriter(d1);
      await writer.write([
        new Document()..append("id", 'id-a').append("all","x"),
        new Document()..append("id", 'id-b').append("all","x")]);

      writer = new IndexWriter(d2);
      await writer.write([
        new Document().append("id", 'id-c').append("all","x"),
        new Document().append("id", 'id-d').append("all","x"),
        new Document().append("id", 'id-e').append("all","x")]);

      await writer.delete(new BoolQuery().append("id", 'id-d'));

      CompositeIndexDirectory compositeDir = new CompositeIndexDirectory([d1, d2]);
      await IndexWriter.copyFrom(compositeDir, d3);

      IndexReader reader = await DirectoryReader.open(d3);
      IndexSearcher searcher = new IndexSearcher(reader);
      expect(await searcher.count(new BoolQuery().append("all", "x")), 4);
    });

    test('test update a document', () async {
      IndexDirectory segment1 = d1;
      IndexWriter writer = new IndexWriter(segment1);
      await writer.write([
        new Document().append("id", '1').append("str","x"),
        new Document().append("id", '2').append("str","x"),
        new Document().append("id", '3').append("str","x")]);

      IndexReader reader = await DirectoryReader.open(d1);
      IndexSearcher searcher = new IndexSearcher(reader);
      expect(await searcher.count(new BoolQuery().append("str", "x")), 3);

      await writer.delete(new BoolQuery().append("id", "2")); //TODO excute searches against all segments
      IndexDirectory segment2 = d2;
      writer = new IndexWriter(segment2);
      await writer.write([
        new Document()..append("id", '2').append("str","y")]);

      CompositeIndexDirectory compositeDir = new CompositeIndexDirectory([d1, d2]);
      reader = await DirectoryReader.open(compositeDir);
      searcher = new IndexSearcher(reader);
      TopDocs hits = await searcher.search(new BoolQuery().append("str", "x"), 10);
      expect(hits.totalHits, 2);
      expect((await searcher.doc(hits.scoreDocs[0].doc)).get("id"), "1");
      expect((await searcher.doc(hits.scoreDocs[1].doc)).get("id"), "3");
      hits = await searcher.search(new BoolQuery().append("str", "y"), 10);
      expect(hits.totalHits, 1);
      expect((await searcher.doc(hits.scoreDocs[0].doc)).get("id"), "2");
    });

  });
}

